/*
 *		Nano inet daemon. :-)
 *
 *	Copyright (c) 2002 by Andy Pershin <apa@puug.org.ru>
 *
 *	This program is free software; you can redistribute it
 *	and/or  modify it under  the terms of  the GNU General
 *	Public  License as  published  by  the  Free  Software
 *	Foundation;  either  version 2 of the License, or  (at
 *	your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY  or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 * $Id: inetd.c,v 1.7 2002/09/15 17:02:01 apa Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <netdb.h>
#include <pwd.h>
#include <grp.h>
#include <signal.h>
#include <errno.h>

#include "gwbox.h"

#define CONFIG "/etc/inetd.conf"

#define ARGS 20

static char **cmd;

struct servtab {
    int sock;
    unsigned short port;
    int socktype;
    int proto;
    char nowait;
    uid_t user;
    gid_t group;
    unsigned char *prog;
    unsigned char *args[ARGS];
    struct servtab *next;
};

static struct servtab *s = NULL;

static void chop(char *str) {

    while (*str) {
        if (*str == '\n' || *str == '\r') {
            *str = '\0';
            break;
        }
        str++;
    }
}

static int getsocktype(const char *socktype) {

    if (strcmp(socktype, "stream") == 0) return SOCK_STREAM;
    else if (strcmp(socktype, "dgram") == 0) return SOCK_DGRAM;
    else if (strcmp(socktype, "raw") == 0) return SOCK_RAW;
    else if (strcmp(socktype, "rdm") == 0) return SOCK_RDM;
    else if (strcmp(socktype, "seqpacket") == 0) return SOCK_SEQPACKET;
    else return -1;
}

static int freeservtab(void) {

    struct servtab *s1, *s2;
    int i;

    s2 = NULL;

    for (s1 = s; s1 != NULL; s1 = s2) {
        s2 = s1->next;
        if (s1->sock >= 0) close(s1->sock);
        s1->sock = -1;
        for (i=0; i < ARGS; i++) {
            if (s1->args[i]) free(s1->args[i]);
        }
        free(s1->prog);
        free(s1);
    }
    return 0;
}

static char **parse(char *buf) {

    int i = 0;
    char *b1, *b2;
    char **str = NULL;

    b1 = buf;

    for (;;) {
        b2 = b1;
        if (*b2 == '\0') break;
        while (*b2 && *b2 != ' ' && *b2 != '\t') b2++;
        *b2++ = '\0';
        str = (char **) realloc(str, (i+1)*sizeof(char *));
        *(str+i) = b1;
        while (*b2 && (*b2 == ' ' || *b2 == '\t')) b2++;
        b1 = b2;
        i++;
    }
    str = (char **) realloc(str, (i+1)*sizeof(char *));
    *(str+i) = NULL;
    return str;
}

static int readconfig(char *file) {

    FILE *fp;
    char buf[256];
    struct servtab *s1, *s2;
    int row = 0;
    char **str;
    struct servent *se;
    struct protoent *pe;
    struct passwd *p;
    struct group *g;
    int i;
    char *b;

    s = s1 = s2 = NULL;

    if ((fp = fopen(file, "r")) == NULL) {
        perror(file);
        return -1;
    }
    while (fgets(buf, sizeof(buf)-1, fp) != NULL) {
        row++;
        if (buf[0] == '#' || buf[0] == '\n' ||
                buf[0] == ' ' || buf[0] == '\t') continue;
        chop(buf);
        s1 = malloc(sizeof(struct servtab));
        if (s1 == NULL) {
            perror("malloc()");
            return -1;
        }
        memset(s1, 0, sizeof(struct servtab));
        if (!s) s = s1;
        if (s2) s2->next = s1;
        s1->next = NULL;
        str = parse(buf);
        for (i = 0; i < 6; i++) {
            if (str[i] == NULL) {
                fprintf(stderr, "Syntax error on string %d (%d)\n", row, i);
                if (str) free(str);
                return -1;
            }
        }
        s1->sock = -1;
	if ((pe = getprotobyname(str[2])) == NULL) {
	    fprintf(stderr, "Unknown protocol: %s\n", str[2]);
            free(str);
            freeservtab();
            return -1;
	}
	s1->proto = pe->p_proto;
        if ((se = getservbyname(str[0], str[2])) == NULL) {
            fprintf(stderr, "Unknown service: %s/%s\n", str[0], str[2]);
            free(str);
            freeservtab();
            return -1;
        }
        s1->port = se->s_port;
        if ((s1->socktype = getsocktype(str[1])) < 0) {
            fprintf(stderr, "Wrong socktype: \"%s\"\n", str[1]);
            free(str);
            freeservtab();
            return -1;
        }
        if (strcmp(str[3], "nowait") == 0) s1->nowait = 1;
        if ((b = strchr(str[4], '.')) != NULL) {
            *b++ = '\0';
            if ((g = getgrnam(b)) != NULL) s1->group = g->gr_gid;
        }
        if ((p = getpwnam(str[4])) == NULL) {
            fprintf(stderr, "Unknown user: \"%s\"\n", str[4]);
            free(str);
            freeservtab();
            return -1;
        }
        s1->user = p->pw_uid;
        if (s->group == 0) s1->group = p->pw_gid;
        s1->prog = strdup(str[5]);
        for (i=0; i < ARGS; i++) {
            if (str[i+6] == NULL) {
                s1->args[i] = NULL;
                break;
            }
            s1->args[i] = strdup(str[i+6]);
        }
        s1->args[i] = NULL;
        free(str);
        s2 = s1;
    }
    s1->next = NULL;
    fclose(fp);
    return 0;
}

static void quit(const int sig) {

    switch (sig) {
        case SIGCHLD:
            waitpid(-1, NULL, WNOHANG);
            break;
        case SIGHUP:
            freeservtab();
            execv(cmd[0], cmd);
            break;
        default:
            log(cmd[0], LOG_NOTICE, "Got signal %d. Exit.", sig);
            exit(sig);
    }
}

static int sockinit(struct servtab *ss) {

    int sock;
    struct sockaddr_in addr;
    int i;

    memset(&addr, 0, sizeof(addr));
    addr.sin_family = AF_INET;
    addr.sin_addr.s_addr = INADDR_ANY;
    addr.sin_port = ss->port;

    if ((sock = socket(PF_INET, ss->socktype, ss->proto)) < 0) {
        log(cmd[0], LOG_ERR, "socket(%s): %s", ss->args[0], strerror(errno));
        return -1;
    }
    i = 1;
    if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &i, sizeof(i)) < 0) {
        log(cmd[0], LOG_ERR, "REUSEADDR(%s): %s", ss->args[0], strerror(errno));
        return -1;
    }
    if (bind(sock, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
        log(cmd[0], LOG_ERR, "bind(%s): %s", ss->args[0], strerror(errno));
        return -1;
    }
    if (listen(sock, 5) < 0) {
        log(cmd[0], LOG_ERR, "listen(%s): %s", ss->args[0], strerror(errno));
        return -1;
    }
    return sock;
}

static int startproc(int sock, struct servtab *ss) {

    switch (fork()) {
        case -1:
            log(cmd[0], LOG_ERR, "Can't fork(): %s", strerror(errno));
            return -1;
        case 0:
            break;
        default:
            return 0;
    }
    if (ss->group) setregid(ss->group, ss->group);
    if (ss->user) setreuid(ss->user, ss->user);
    close(STDIN_FILENO);
    close(STDOUT_FILENO);
    close(STDERR_FILENO);
    dup2(sock, STDIN_FILENO);
    dup2(sock, STDOUT_FILENO);
    dup2(sock, STDERR_FILENO);
    if (sock > STDERR_FILENO) close(sock);
    execv(ss->prog, (char **const)ss->args);
    log(cmd[0], LOG_ERR, "Can't exec %s: %s", ss->prog, strerror(errno));
    return -1;
}

static int inetd(void) {

    struct servtab *s1;
    int maxfd = 0;
    fd_set fds;
    struct timeval tv;
    int sock;
    int ret;
    struct sockaddr_in from;
    socklen_t fromlen;

    for (s1 = s; s1 != NULL; s1 = s1->next) {
        if ((s1->sock = sockinit(s1)) < 0) continue;
        maxfd = (maxfd > s1->sock) ? maxfd : s1->sock;
    }
    FD_ZERO(&fds);
    for (;;) {
        for (s1 = s; s1 != NULL; s1 = s1->next) {
            if (s1->sock >= 0) FD_SET(s1->sock, &fds);
        }
        tv.tv_sec = 10;
        tv.tv_usec = 0;
        ret = select(maxfd+1, &fds, NULL, NULL, &tv);
        if (ret == 0) continue;
        for (s1 = s; s1 != NULL; s1 = s1->next) {
            if (s1->sock >= 0 && FD_ISSET(s1->sock, &fds)) {
                fromlen = sizeof(from);
                if ((sock = accept(s1->sock, (struct sockaddr *)&from, &fromlen)) < 0) {
                    log(cmd[0], LOG_ERR, "accept(): %s", strerror(errno));
                    continue;
                }
                startproc(sock, s1);
                close(sock);
                FD_CLR(s1->sock, &fds);
            }
        }
    }
    freeservtab();
    return 0;
}

int inetd_main (int argc, char *argv[]) {

    int i;
    int bg = 1;                 /* default: background */

    cmd = argv;

    while ((i = getopt(argc, argv, "i")) != EOF) {
        switch (i) {
            case 'i':
                bg = 0;
                break;
            default:
                fprintf(stderr, "Usage: %s [-i] [config]\n", argv[0]);
                return 1;
        }
    }
    if (readconfig(argv[optind]?argv[optind]:CONFIG) < 0) return 2;

    if (bg) {
        if (make_daemon(argv[0], &quit) < 0) return 2;
    }
    return inetd();
}
