/*
 *		support routines for gwbox
 *
 *	Copyright (c) 2002 by Andy Pershin <apa@puug.org.ru>
 *
 *	This program is free software; you can redistribute it
 *	and/or  modify it under  the terms of  the GNU General
 *	Public  License as  published  by  the  Free  Software
 *	Foundation;  either  version 2 of the License, or  (at
 *	your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY  or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 * $Id: misc.c,v 1.8 2002/10/03 13:03:18 apa Exp $
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <stdarg.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <utmp.h>

#include "gwbox.h"

void log(const char *prog, int level, char *fmt, ...) {

    va_list ap;
    char buf[256];

    va_start(ap, fmt);
    vsnprintf(buf, sizeof(buf), fmt, ap);
#ifdef LOG_PERROR
    openlog(prog, LOG_PID | LOG_PERROR, LOG_DAEMON);
#else
    openlog(prog, LOG_PID, LOG_DAEMON);
    fprintf(stderr, "%s[%d]: %s\n", prog, getpid(), buf);
#endif
    syslog(level, buf);
    closelog();
    va_end(ap); 
}

void issue(int fd) {

    FILE *fp;
    unsigned char buf[256];
    int ret;

    if (write(fd, "\r\n", 2) <= 0) return;
    if ((fp = fopen("/etc/issue.net", "r")) == NULL) {
        if ((fp = fopen("/etc/issue", "r")) == NULL) return;
    }
    while(fgets(buf, sizeof(buf)-2, fp)) {
        buf[strlen(buf)-1] = '\0';	/* chop; :-) */
        ret = sprintf(buf, "%s\r\n", buf);
        if (write(fd, buf, ret) <= 0) break;
    }
    fclose(fp);
}

int sock2fd(int sock, int fd, int timeout, const char *prog,
			int (*filter1)(unsigned char *buf, int len),
			int (*filter2)(unsigned char *buf, int len)) {

    fd_set fds;
    struct timeval tv;
    int ret;
    unsigned char buf[2048];
    int maxfd = (sock>fd)?sock:fd;

    FD_ZERO(&fds);
    for (;;) {
        FD_SET(sock, &fds);
        FD_SET(fd, &fds);
        tv.tv_sec = timeout;          /* inactivity timer */
        tv.tv_usec = 0;
        ret = select(maxfd+1, &fds, NULL, NULL, &tv);
        if (ret == 0) {
            log(prog, LOG_NOTICE, "Connection timed out...");
            break;
        }
        if (FD_ISSET(sock, &fds)) {
            if ((ret = read(sock, buf, sizeof(buf)-10)) <= 0) {
                break;
            }
            if(filter1) ret = filter1(buf, ret);
            if (ret && write(fd, buf, ret) <= 0) {
                break;
            }
            FD_CLR(sock, &fds);
        } else if (FD_ISSET(fd, &fds)) {
            if ((ret = read(fd, buf, sizeof(buf)-10)) <= 0) {
                break;
            }
	    if(filter2) ret = filter2(buf, ret);
            if (ret && write(sock, buf, ret) <= 0) {
                break;
            }
            FD_CLR(fd, &fds);
        }
    }
    return 0;
}

void show_x25_route(void) {

    FILE *fp;
    char buf[128];
    char *p;
    int digits;

    if ((fp = fopen("/proc/net/x25_routes", "r")) == NULL) return;

    fputs("Kernel X.25 routing table\n", stdout);
    fputs("Destination          Interface\n", stdout);
    fgets(buf, sizeof(buf), fp);	/* <-- skip header */

    while (fgets(buf, sizeof(buf), fp) != NULL) {
        p = strchr(buf, '\n');
        if (p) *p = 0;
        buf[24] = 0;
        buf[35] = 0;
        digits = atoi(buf+17);
        if (digits < 0 || digits > 15) digits = 15;
        buf[digits] = 0;
        if (digits == 0) {
            printf("*                    %-5s\n", buf+25);
        } else {
            printf("%s/%*d   %-5s\n", buf, digits-17, digits, buf+25);
	    }
    }
    fclose(fp);
}

int make_daemon(const char *prog, void (*handler)(const int sig)) {

    int fd;
    struct sigaction sa;

    if (fork() != 0) exit(0);

    memset(&sa, 0, sizeof(sa));
    sa.sa_flags |= SA_NOCLDSTOP;
    sa.sa_flags |= SA_RESTART;
    sa.sa_handler = handler;
    sigaction(SIGHUP, &sa, NULL);
    sigaction(SIGTERM, &sa, NULL);
    sigaction(SIGINT, &sa, NULL);
    sigaction(SIGQUIT, &sa, NULL);
    sigaction(SIGCHLD, &sa, NULL);

    /*  stdin, stdout  stderr  /dev/null */
    if ((fd = open("/dev/null", O_RDWR)) < 0) {
        log(prog, LOG_ERR, "open(/dev/null): %s", strerror(errno));
        return -1;
    }
    close(STDIN_FILENO);
    close(STDOUT_FILENO);
    close(STDERR_FILENO);
    if (dup2(fd, STDIN_FILENO) != STDIN_FILENO) {
        log(prog, LOG_ERR, "dup2(stdin): %s", strerror(errno));
        close(fd);
        return -1;
    }
    if (dup2(fd, STDOUT_FILENO) != STDOUT_FILENO) {
        log(prog, LOG_ERR, "dup2(stdout): %s", strerror(errno));
        close(fd);
        return -1;
    }
    if (dup2(fd, STDERR_FILENO) != STDERR_FILENO) {
        log(prog, LOG_ERR, "dup2(stderr): %s", strerror(errno));
        close(fd);
        return -1;
    }
    if (fd > STDERR_FILENO) {
        close(fd);
    }
    setsid();
    sa.sa_handler = SIG_IGN;
    sigaction(SIGTTOU, &sa, NULL);
    sigaction(SIGTTIN, &sa, NULL);
    sigaction(SIGTSTP, &sa, NULL);
    chdir("/");
    return 0;
}

void clearutmp(const char *tty) {

    struct utmp *u;

    setutent();
    while ((u = getutent())) {
	if (u->ut_type == USER_PROCESS &&
	    strcmp(u->ut_line, tty) == 0) {
	    u->ut_type = DEAD_PROCESS;
	    memset(u->ut_host, 0, sizeof(u->ut_host));
	    memset(u->ut_user, 0, sizeof(u->ut_user));
	    pututline(u);
	    break;
	}
    }
    endutent();
}
