/*
 *	netstat - print network statistics (connections, routing table)
 *
 *	Copyright (c) 2002 by Andy Pershin <apa@puug.org.ru>
 *
 *	This program is free software; you can redistribute it
 *	and/or  modify it under  the terms of  the GNU General
 *	Public  License as  published  by  the  Free  Software
 *	Foundation;  either  version 2 of the License, or  (at
 *	your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY  or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 * $Id: netstat.c,v 1.3 2002/09/11 13:35:39 apa Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <net/route.h>

#include "gwbox.h"

#define PROCNET "/proc/net"

static void show_ip_route(void) {

    char buf[256];
    char dev[IFNAMSIZ];
    struct sockaddr_in dst, gw, mask;
    int flgs, ref, use, metric, mtu, win, irtt;
    char flags[64];
    FILE *fp;
    char dest[16], gateway[16], genmask[16];
    int i;

    if ((fp = fopen(PROCNET "/route", "r")) == NULL) return;

    printf("Kernel IP routing table\n");
    printf("Destination     Gateway         Genmask         Flags Metric Win Irtt Iface\n");
    fgets(buf, sizeof(buf), fp);
    while (fgets(buf, sizeof(buf), fp) != NULL) {
	i = 0;
	while (buf[i] != ' ' && buf[i] != '\n') i++;
	buf[i] = '\0';
	sscanf(buf, "%s\t%X\t%X\t%d\t%d\t%d\t%d\t%X\t%X\t%d\t%d",
			dev, &dst.sin_addr.s_addr, &gw.sin_addr.s_addr,
			&flgs, &ref, &use, &metric, &mask.sin_addr.s_addr,
			&mtu, &win, &irtt);
	i = 0;
	if (flgs & RTF_UP) {
	    if (flgs & RTF_REJECT) flags[i++] = '!';
	    else flags[i++] = 'U';
	    if (flgs & RTF_GATEWAY) flags[i++] = 'G';
	    if (flgs & RTF_HOST) flags[i++] = 'H';
	    if (flgs & RTF_REINSTATE) flags[i++] = 'R';
	    if (flgs & RTF_DYNAMIC) flags[i++] = 'D';
	    if (flgs & RTF_MODIFIED) flags[i++] = 'M';
	    flags[i] = 0;
	}
	sprintf(dest, "%s", inet_ntoa(dst.sin_addr));
	sprintf(gateway, "%s", inet_ntoa(gw.sin_addr));
	sprintf(genmask, "%s", inet_ntoa(mask.sin_addr));
	printf("%-16s%-16s%-16s%-6s%-6d %-3d %4d %s\n",
		dest, gateway, genmask, flags, metric, win, irtt, dev);
    }
}

static void show_ip_state(const char *proto) {

    char buf[256];
    FILE *fp;
    int d;
    char locaddr[25], remaddr[25];
    unsigned int lport, rport;
    int state, timer_run, uid, timeout;
    long unsigned int txq, rxq, time_len, retr, inode;
    char more[512];
    struct sockaddr_in laddr, raddr;
    const char *tcp_state[] = {
	"",
	"ESTABLISHED",
	"SYN_SENT",
	"SYN_RECV",
	"FIN_WAIT1",
	"FIN_WAIT2",
	"TIME_WAIT",
	"CLOSE",
	"CLOSE_WAIT",
	"LAST_ACK",
	"LISTEN",
	"CLOSING"
    };

    sprintf(buf, "%s/%s", PROCNET, proto);
    if ((fp = fopen(buf, "r")) == NULL) return;
    while (fgets(buf, sizeof(buf), fp) != NULL) {
        if (strchr(buf, ':') == NULL) continue;
        sscanf(buf, "%d: %X:%X %X:%X %X %lX:%lX %X:%lX %lX %d %d %ld %512s\n",
                &d, &laddr.sin_addr.s_addr, &lport, &raddr.sin_addr.s_addr, &rport, &state, &txq, &rxq,
                &timer_run, &time_len, &retr, &uid, &timeout, &inode, more);
        sprintf(locaddr, "%s:%d", inet_ntoa(laddr.sin_addr), lport);
        sprintf(remaddr, "%s:%d", inet_ntoa(raddr.sin_addr), rport);
        printf("%-5s %6ld %6ld %-23s %-23s %-12s\n", proto, rxq, txq, locaddr, remaddr,
                tcp_state[state]);
	fflush(stdout);
    }
    fclose(fp);
}

static void show_x25_state(void) {

    char buf[256];
    char dev[16];
    FILE *fp;
    int st, vs, vr, sendq, recvq, lci;
    static char *x25_state[] = {
	"LISTENING",
	"SABM_SENT",
	"DISC_SENT",
	"ESTABLISHED",
	"RECOVERY"
    };

    sprintf(buf, "%s/%s", PROCNET, "x25");
    if ((fp = fopen(buf, "r")) == NULL) return;
    printf("Active X.25 sockets\n");
    printf("Dest            Source          Device  LCI  State        Vr/Vs  Send-Q  Recv-Q\n");
    fgets(buf, sizeof(buf)-1, fp);
    while(fgets(buf, sizeof(buf)-1, fp)) {
	buf[10] = 0;
	buf[20] = 0;
	sscanf(buf+22,"%s %d %d %d %d %*d %*d %*d %*d %*d %*d %d %d %*d",
			dev, &lci, &st, &vs, &vr, &sendq, &recvq);
	printf("%-15s %-15s %-7s %-3d  %-11s  %02d/%02d  %-6d  %-6d\n",
		buf, buf+11, dev, lci, x25_state[st], vr, vs, sendq, recvq);
    }
    fclose(fp);
    return;
}

int netstat_main (int argc, char *argv[]) {

    int i;

    while((i = getopt(argc, argv, "rhaln")) != EOF) {
	switch (i) {
	    case 'r':
		show_ip_route();
		show_x25_route();
		return 0;
	    case 'h':
		usage(argv[0]);
		return 0;
	    default:
		break;
	}
    }
    fputs("Active Internet connections\n", stdout);
    fputs("Proto Recv-Q Send-Q Local Address           Foreign Address         State\n", stdout);
    show_ip_state("tcp");
    show_ip_state("udp");
    show_x25_state();
    return 0;
}
