/* 
 *                   Simple telnet server
 *
 *	Copyright (c) 2002 by Andy Pershin <apa@puug.org.ru>
 *
 *	This program is free software; you can redistribute it
 *	and/or  modify it under  the terms of  the GNU General
 *	Public  License as  published  by  the  Free  Software
 *	Foundation;  either  version 2 of the License, or  (at
 *	your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY  or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 * $Id: telnetd.c,v 1.5 2002/10/03 13:03:18 apa Exp $
 *
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <termios.h>
#include <arpa/telnet.h>
#include <pty.h>

#include "gwbox.h"

#define BUFSIZE 4000

#define MIN(a,b) ((a) > (b) ? (b) : (a))

static char *argv0;
static char transparent = 0;

/* taken from utelnetd. */
static int rmiac(unsigned char *buf, int len) {

    unsigned char *in = buf;
    unsigned char *out = buf;
    unsigned char *end = buf + len;
    int i = 0;

    while (in < end) {
	if (*in != IAC) {
	    *out++ = *in++;
	    i++;
	} else {
	    if ((in+2) < end) {
		/* the entire IAC is contained in the buffer 
		   we were asked to process. */
		in += 3;
	    } else {
		/* only the beginning of the IAC is in the 
		   buffer we were asked to process, we can't
		   process this char. */
		break;
	    }
	}
    }
    return i;
}

int telnetd_main(int argc, char *argv[]) {

    int c;
    char buf[256];
    int ret;
    struct sockaddr_in from;
    pid_t pid;
    int pty;
    char *tty = NULL;
    char *login = "/bin/login";
    int timeout = 300;			/* 5 minutes... */
    struct termios tp;

    argv0 = argv[0];

    while ((c = getopt(argc, argv, "tl:w:h")) != EOF) {
	switch (c) {
	    case 't':
		transparent = 1;
		break;
	    case 'l':
		login = optarg;
		break;
	    case 'w':
		timeout = atoi(optarg);
		break;
	    case 'h': 
		default:
		usage(argv[0]);
		return 1;
	}
    }

    if (access(login, X_OK) < 0) {
	log(argv0, LOG_ERR, "%s: %s", login, strerror(errno));
	return 2;
    }

    ret = sizeof(from);
    if (getpeername(STDIN_FILENO, (struct sockaddr *)&from, &ret) < 0) {
	log(argv0, LOG_ERR, "getpeername(): %s", strerror(errno));
	return 3;
    }
    ret = 1;
    if (setsockopt(STDIN_FILENO, SOL_SOCKET, SO_KEEPALIVE, &ret, sizeof (ret)) < 0) {
	log(argv0, LOG_ERR, "KEEPALIVE: %s", strerror(errno));
	return 3;
    }

    /* send iac commands */
    if (!transparent) {
	c = 0;
	buf[c++] = IAC;
	buf[c++] = DO;
	buf[c++] = TELOPT_ECHO;
	buf[c++] = IAC;
	buf[c++] = DO;
	buf[c++] = TELOPT_LFLOW;
	buf[c++] = IAC;
	buf[c++] = WILL;
	buf[c++] = TELOPT_ECHO;
	buf[c++] = IAC;
	buf[c++] = WILL;
	buf[c++] = TELOPT_SGA;
	if (write(STDOUT_FILENO, buf, c) <= 0) {
	    log(argv0, LOG_ERR, "write(): %s", strerror(errno));
	    return 5;
	}
	/* Print issue (NO issue.net(5) support!) */
	issue(0);
    }

    pid = forkpty(&pty, tty, NULL, NULL);
    if (pid == -1) {
	log(argv0, LOG_ERR, "forkpty(): %s", strerror(errno));
	return 4;
    } else if (pid == 0) {
	/* child */
	if (!transparent) {
	    tcgetattr(STDIN_FILENO, &tp);
	    tp.c_lflag |= ECHO;
	    tp.c_lflag &= ~ECHOCTL;
	    tp.c_oflag |= (OPOST | ONLCR | XTABS);
	    //tp.c_oflag |= (ONLCR | XTABS);
	    tp.c_iflag |= ICRNL;
	    tp.c_iflag &= ~IXOFF;
	    tp.c_cflag |= CS8;
	    /* termbuf.c_lflag &= ~ICANON; */
	    tcsetattr(STDIN_FILENO, TCSANOW, &tp);
        }
	execl(login, "login", "-h", inet_ntoa(from.sin_addr), NULL);
        log(argv0, LOG_ERR, "%s: %s", login, strerror(errno));
        return 4;
    }

    if ((ret = open("/dev/null", O_RDWR)) < 0) {
	log(argv0, LOG_ERR, "open(/dev/null): %s", strerror(errno));
	return 4;
    }
    close(STDERR_FILENO);
    if (dup2(ret, STDERR_FILENO) != STDERR_FILENO) {
        log(argv0, LOG_ERR, "dup2(stderr): %s", strerror(errno));
        close(ret);
        return 4;
    }
    if (ret > STDERR_FILENO) {
        close(ret);
    }
    log(argv0, LOG_INFO, "Connect from %s:%d",
	    inet_ntoa(from.sin_addr), ntohs(from.sin_port));
    ret = sock2fd(0, pty, timeout, argv0, (transparent)?NULL:&rmiac, NULL);
    log(argv0, LOG_INFO, "Connect from %s:%d closed",
	    inet_ntoa(from.sin_addr), ntohs(from.sin_port));
    clearutmp(tty);
    close(pty);
    return ret;
}
