/*
 *			X.25 Server :-)
 *
 *	Usage: x25d -d dev <X.121 ADDRESS>
 *
 *	Copyright (c) 2002 by Andy Pershin <apa@puug.org.ru>
 *
 *	This program is free software; you can redistribute it
 *	and/or  modify it under  the terms of  the GNU General
 *	Public  License as  published  by  the  Free  Software
 *	Foundation;  either  version 2 of the License, or  (at
 *	your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY  or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 * $Id: x25d.c,v 1.2 2002/09/12 09:00:51 apa Exp $
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/socket.h>
#include <signal.h>
#include <errno.h>
#include <net/if.h>
#include <linux/x25.h>
#include <linux/if_tun.h>
#include <pty.h>

#include "gwbox.h"

#define CUD_PAD 0x01
#define CUD_IP  0xCC

static char *argv0;
static int timeout = 300;

static void quit(const int sig) {

    if (sig == SIGCHLD) {
        waitpid(-1, NULL, WNOHANG);
	return;
    }
    log(argv0, LOG_NOTICE, "Got signal %d. Exit.", sig);
    exit(sig);
}

static int addtunbytes(unsigned char *buf, int len) {

    memmove(buf+4, buf, len);
    buf[0] = 0x00;
    buf[1] = 0x00;
    buf[2] = 0x08;
    buf[3] = 0x00;
    return (len+4);
}

static int remtunbytes(unsigned char *buf, int len) {

    len -= 4;
    memmove(buf, buf+4, len);
    return len;
}

static int make_tundev(char *device) {

    int fd;
    struct ifreq ifr;

    if((fd = open("/dev/net/tun", O_RDWR)) < 0) {
	log(argv0, LOG_ERR, "/dev/net/tun: %s", strerror(errno));
	return -1;
    }
    ifr.ifr_flags = IFF_TUN;
    strcpy(ifr.ifr_name, device);
    if(ioctl(fd, TUNSETIFF, (void *) &ifr) < 0) {
	log(argv0, LOG_ERR, "TUNSETIFF: %s", strerror(errno));
	close(fd);
	return -1;
    }
    strcpy(device, ifr.ifr_name);
    log(argv0, LOG_INFO, "Allocating \"%s\" device.", device);
    return fd;
}

static int start_pad(int sock, struct sockaddr_x25 *sx25) {

    int pty;
    pid_t pid;
    unsigned char buf[256];
    int ret;

    issue(sock);

    pid = forkpty(&pty, NULL, NULL, NULL);
    if (pid == -1) {
	log(argv0, LOG_ERR, "forkpty(): %s", strerror(errno));
	return -1;
    } else if (pid == 0) {
        /* child */
	close(sock);
	sprintf(buf, "x25-%s", sx25->sx25_addr.x25_addr);
        execl("/bin/login", "login", "-h", buf, NULL);
        log(argv0, LOG_ERR, "login: %s", strerror(errno));
        return -1;
    }
    ret = sock2fd(sock, pty, timeout, argv0, NULL, NULL);
    close(pty);
    return ret;
}

static int x25d(int tunfd, char *addr) {

    int sock, sock2;
    struct sockaddr_x25 sx25, fromx25;
    struct x25_calluserdata cud;
    int i;

    memset(&sx25, 0, sizeof(sx25));
    memset(&fromx25, 0, sizeof(sx25));
    sx25.sx25_family = AF_X25;
    strncpy(sx25.sx25_addr.x25_addr, addr, sizeof(sx25.sx25_addr.x25_addr)-1);

    if ((sock = socket(AF_X25, SOCK_SEQPACKET, 0)) < 0) {
	log(argv0, LOG_ERR, "socket(): %s", strerror(errno));
	return 2;
    }
    i = 1;
    if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &i, sizeof(i)) < 0) {
	log(argv0, LOG_ERR, "REUSEADDR: %s", strerror(errno));
	return 2;
    }
    if (bind(sock, (struct sockaddr *)&sx25, sizeof(sx25)) < 0) {
	log(argv0, LOG_ERR, "bind(): %s", strerror(errno));
	return 2;
    }
    if (listen(sock, 1) < 0) {
	log(argv0, LOG_ERR, "listen(): %s", strerror(errno));
	return 2;
    }
    /*   ! */
    make_daemon(argv0, &quit);
    for (;;) {
	i = sizeof(sx25);
	if ((sock2 = accept(sock, (struct sockaddr *)&sx25, &i)) < 0) {
	    log(argv0, LOG_ERR, "accept(): %s", strerror(errno));
	    return 2;
	}
	i = fork();
	if (i == -1) {
	    log(argv0, LOG_ERR, "fork(): %s", strerror(errno));
	} else if (i != 0) {
	    close(sock2);
	    continue;
	}
	/* child */
	close(sock);
	i = sizeof(fromx25);
	if (getpeername(sock2, (struct sockaddr *)&fromx25, &i) < 0) {
	    log(argv0, LOG_ERR, "getpeername(): %s", strerror(errno));
	    return 2;
	}
	log(argv0, LOG_INFO, "Connect from \"%s\"", fromx25.sx25_addr.x25_addr);
	/* Call User Data */
	if (ioctl(sock2, SIOCX25GCALLUSERDATA, &cud) < 0) {
	    log(argv0, LOG_ERR, "SIOCX25GCALLUSERDATA: %s", strerror(errno));
	    return 2;
	}
	switch (cud.cuddata[0]) {
	    case CUD_PAD:
		log(argv0, LOG_INFO, "Opening PAD session with %s", fromx25.sx25_addr.x25_addr);
		start_pad(sock2, &fromx25);
		break;
	    case CUD_IP:
		if (tunfd < 0) {
		    log(argv0, LOG_WARNING, "No TUN device available.");
		    break;
		}
		log(argv0, LOG_INFO, "Opening IP session with %s", fromx25.sx25_addr.x25_addr);
		sock2fd(sock2, tunfd, timeout, argv0, &addtunbytes, &remtunbytes);
		break;
	    default:
		/* Unknown encapsulation... */
		log(argv0, LOG_WARNING, "Unknown CUD: 0x%02X -- refusing.", cud.cuddata[0]);
		break;
	}
	log(argv0, LOG_INFO, "Connection from \"%s\" closed", fromx25.sx25_addr.x25_addr);
	exit(0);
    }
    return 0;
}

int x25d_main (int argc, char *argv[]) {

    int i;
    char device[IFNAMSIZ];
    int tunnel = -1;

    if (argc < 2) {
	usage(argv[0]);
	return 1;
    }

    argv0 = argv[0];

    memset(device, 0, sizeof(device));
    while ((i = getopt(argc, argv, "hd:t:")) != EOF) {
	switch(i) {
	    case 'd':
		strncpy(device, optarg, sizeof(device)-1);
		break;
	    case 't':
		timeout = atoi(optarg);
		break;
	    case 'h':
	    default:
		usage(argv[0]);
		return 1;
	}
    }
    if (optind >= argc) {
	usage(argv[0]);
	return 1;
    }
    tunnel = make_tundev(device);
    return x25d(tunnel, argv[optind]);
}
